/**
* The contents of this file are subject to the Mozilla Public 
* License Version 1.1 (the "License"); you may not use this 
* file except in compliance with the License. You may obtain a
* copy of the License at http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an 
* "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express 
* or implied. See the License for the specific language 
* governing rights and limitations under the License.
*
* The Original Code is this file as it was released on
* May 30, 2005.
*
* The Initial Developer of the Original Code is Terry Yuen. 
* Portions created by Terry Yuen are Copyright (C) 2005.
* All Rights Reserved.
**/

const Highlighter = {
  toggleMenuState: false,
  autoHighlightMode: false,
  shouldDisplayIcons: false,
  warnBeforeLeave: false,
  lastColors: {fgcolor:"", bgcolor:""},

  startup: function() {
    document.getElementById("contentAreaContextMenu")
        .addEventListener("popupshowing", Highlighter.popupshowing, false);
    gBrowser.addEventListener("mouseup", Highlighter.autoHighlight, false);
    gBrowser.addEventListener("mousedown", Highlighter.trackKeyStates, false);
    gBrowser.addEventListener("keydown", Highlighter.trackKeyStates, false);

    Highlighter.loadSettings();
  },

  destroy: function() {
    window.removeEventListener("load", Highlighter.startup, false);
    document.getElementById("contentAreaContextMenu")
        .removeEventListener("popupshowing", Highlighter.popupshowing, false);
    gBrowser.removeEventListener("mouseup", Highlighter.autoHighlight, false);
    gBrowser.removeEventListener("mousedown", Highlighter.trackKeyStates, false);
    gBrowser.removeEventListener("keydown", Highlighter.trackKeyStates, false);
  },

  loadSettings: function() {
    const pref = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("highlighter.");

    Highlighter.shouldDisplayIcons = pref.getBoolPref("display_icons");
    Highlighter.warnBeforeLeave = pref.getBoolPref("warn_leave");

    var arr = pref.getCharPref("colours").split("|");
    Highlighter.lastColors.bgcolor = arr[0];
    Highlighter.lastColors.fgcolor = arr[1];

    if(arr.length % 2 != 0)throw "Highlighter: number of colours not an even number.";

    var n = document.getElementById("context-highlighter-popup").firstChild;
    for(var i=0; i<arr.length && n;) {
      document.getAnonymousElementByAttribute(n, "class", "menu-iconic-icon").style.backgroundColor = arr[i];
      n.setAttribute("bgcolor", arr[i++]);
      n.setAttribute("fgcolor", arr[i++]);
      n = n.nextSibling;
    }

    var n = document.getElementById("highlighterMenu").firstChild;
    for(var i=0; i<arr.length && n;) {
      document.getAnonymousElementByAttribute(n, "class", "menu-iconic-icon").style.backgroundColor = arr[i];
      n.setAttribute("bgcolor", arr[i++]);
      n.setAttribute("fgcolor", arr[i++]);
      n = n.nextSibling;
    }
  },

  popupshowing: function(event) {
    var n = document.popupNode;
    var sel = n.ownerDocument.defaultView.getSelection();
    var hasText = (sel.rangeCount == 0) ? false : (sel.getRangeAt(0).collapsed ? false : true);
    var submenu = !gContextMenu.onTextInput && (Highlighter.toggleMenuState || Highlighter.autoHighlightMode);

    gContextMenu.showItem("context-highlighter-submenu", submenu);
    if(hasText) {
      gContextMenu.showItem("context-highlighter-highlight", !submenu);
      gContextMenu.showItem("context-highlighter-sep", false);
      gContextMenu.showItem("context-highlighter-remove", false);
      gContextMenu.showItem("context-highlighter-copy", false);
      gContextMenu.showItem("context-highlighter-instant", Highlighter.autoHighlightMode);
      if(submenu) {
        gContextMenu.showItem("context-highlighter-remove2", false);
        gContextMenu.showItem("context-highlighter-instant2", false);
      }
    } else {
      var isHighlight = (n.nodeType == Node.ELEMENT_NODE && n.hasAttribute("highlighter"));
      gContextMenu.showItem("context-highlighter-highlight", false);
      gContextMenu.showItem("context-highlighter-sep", true);
      gContextMenu.showItem("context-highlighter-remove", isHighlight && !submenu);
      gContextMenu.showItem("context-highlighter-copy", true);
      gContextMenu.showItem("context-highlighter-instant", Highlighter.autoHighlightMode);
      if(submenu) {
        gContextMenu.showItem("context-highlighter-remove2", isHighlight);
        gContextMenu.showItem("context-highlighter-instant2", true);
      }
    }
  },

  trackKeyStates: function(event) {  //always fires before onpopupshowing to capture accesskeys
    Highlighter.toggleMenuState = (event.shiftKey && (event.ctrlKey || event.type == "mousedown"));
  },

  highlightDoc: function(context) {
    var node = document.popupNode;
    var win = node.ownerDocument.defaultView;
    var sel = win.getSelection();

    if(context && context.hasAttribute("bgcolor")) {
      Highlighter.lastColors.fgcolor = context.getAttribute("fgcolor");
      Highlighter.lastColors.bgcolor = context.getAttribute("bgcolor");
    }
    var fgcolor = Highlighter.lastColors.fgcolor;
    var bgcolor = Highlighter.lastColors.bgcolor;

    //if exist highlight and no selections then change current highlight colour
    if((sel.rangeCount == 0 || sel.getRangeAt(0).collapsed)
        && node.nodeType == Node.ELEMENT_NODE && node.hasAttribute("highlighter")) {

      Highlighter.changeHighlightColor(node, fgcolor, bgcolor);

    } else if(sel.rangeCount > 0) {
      Highlighter.highlightRange(win, fgcolor, bgcolor);
    }
  },

  highlightRange: function(win, fgcolor, bgcolor) {
    const record = {offsetY:Number.NaN, firstNode:null, lastNode:null};

    const wrap = win.document.createElement("SPAN");
    wrap.setAttribute("highlighter", bgcolor);
    wrap.style.color = fgcolor;
    wrap.style.backgroundColor = bgcolor;

    const _createWrapper = function(n) {
      var e = wrap.cloneNode(false);

      if(!record.firstNode)record.firstNode = e;
      if(record.lastNode)record.lastNode.nextHighlight = e;
      record.lastNode = e;

      var posTop = win.document.getBoxObjectFor(n.parentNode).y;
      var pageTop = parseInt(win.pageYOffset);
      if(!posTop || posTop < pageTop) {
        record.offsetY = pageTop;
      } else {
        if(!(posTop > record.offsetY))record.offsetY = posTop;
      }

      return e;
    };

    for(var i=0, sel=win.getSelection(); i<sel.rangeCount; i++) {
      Highlighter.highlightRange0(sel.getRangeAt(i), _createWrapper, record);
    }

    if(record.firstNode != null) {
      record.lastNode.nextHighlight = record; //connect linked list back to record

      if(!win.highlighterInfo)win.highlighterInfo = {highlights:[], dirty:true};

      var arr = win.highlighterInfo.highlights;
      //do an insertion sort insertion
      if(arr.length > 0) {
        for(var i=arr.length-1; i>=-1; i--) {
          if(i < 0) {
            arr.splice(0, 0, record);
          } else if(arr[i].offsetY <= record.offsetY) {
            arr.splice(i + 1, 0, record);
            break;
          }
        }
      } else arr.push(record);

      win.highlighterInfo.dirty = true;

      if(Highlighter.warnBeforeLeave)
        Highlighter.ensureUnloadWarning(win);

      if(Highlighter.shouldDisplayIcons)
        Highlighter.ensureHighlighterView(win);
    }
  },

  highlightRange0: function(range, _createWrapper, record) {

    if(range.collapsed)  //(startContainer == endContainer && startOffset == endOffset)
      return;

    var startSide = range.startContainer;
    var endSide = range.endContainer;
    var ancestor = range.commonAncestorContainer;
    var dirIsLeaf = true;

    if(range.endOffset == 0) {  //nodeValue = text | element
      while(!endSide.previousSibling && endSide.parentNode != ancestor) {
        endSide = endSide.parentNode;
      }
      endSide = endSide.previousSibling;
      //dump("\nendOffset=0, endSide="+endSide+" goto root="+endSide+" ancestor="+ancestor);
    } else if(endSide.nodeType == Node.TEXT_NODE) {
      if(range.endOffset < endSide.nodeValue.length) {
        endSide.splitText(range.endOffset);
        //dump("\nendSide is textnode, split text, endSide="+endSide);
      }
    } else if(range.endOffset > 0) {  //nodeValue = element
      endSide = endSide.childNodes.item(range.endOffset - 1);
      //dump("\nendOffset > 0, select indexed="+endSide);
    }

    if(startSide.nodeType == Node.TEXT_NODE) {
      if(range.startOffset == startSide.nodeValue.length) {
        //dump("\nstartOffset=len, noop");
        dirIsLeaf = false;
      } else if(range.startOffset > 0) {
        startSide = startSide.splitText(range.startOffset);
        if(endSide == startSide.previousSibling)endSide = startSide;
        //dump("\nstartOffset > 0, split text, startSide="+startSide);
      }
    } else if(range.startOffset < startSide.childNodes.length) {
      startSide = startSide.childNodes.item(range.startOffset);
      //dump("\nstartOffset < len, select indexed="+startSide);
    } else {
      //dump("\nstartOffset=len, noop");
      dirIsLeaf = false;
    }

    range.setStart(range.startContainer, 0);
    range.setEnd(range.startContainer, 0);

    //dump("\nstartSide="+startSide+"\nendSide="+endSide);
    var done = false;
    var node = startSide;
    var tmp;

    do {
      if(dirIsLeaf && node.nodeType == Node.TEXT_NODE &&
        !((tmp = node.parentNode) instanceof HTMLTableElement ||
          tmp instanceof HTMLTableRowElement ||
          tmp instanceof HTMLTableColElement ||
          tmp instanceof HTMLTableSectionElement)) {
        var wrap = node.previousSibling;
        if(!wrap || wrap != record.lastNode) {
          wrap = _createWrapper(node);
          node.parentNode.insertBefore(wrap, node);
        }
        wrap.appendChild(node);

        node = wrap.lastChild;
        dirIsLeaf = false;
      }

      if(node == endSide && (!endSide.hasChildNodes() || !dirIsLeaf)) {
        done = true;
      }

      if(node instanceof HTMLScriptElement ||
         node instanceof HTMLStyleElement ||
         node instanceof HTMLSelectElement) {  //never parse their children
        dirIsLeaf = false;
      }


      if(dirIsLeaf && node.hasChildNodes()) {
        node = node.firstChild;  //dump("-> firstchild ");

      } else if(node.nextSibling != null) {
        node = node.nextSibling;  //dump("-> nextSibling ");
        dirIsLeaf = true;

      } else if(node.nextSibling == null) {
        node = node.parentNode;  //dump("-> parent ");
        dirIsLeaf = false;
      }
      //if(node == ancestor.parentNode)dump("\nHALT shouldn't face ancestor");
    } while(!done);
  },

  removeHighlight: function() {
    var node = document.popupNode;
    if(node.nodeType == Node.ELEMENT_NODE && node.hasAttribute("highlighter")) {

      var n = node.nextHighlight;
      while(n instanceof HTMLSpanElement && n.hasAttribute("highlighter")) {
        while(n.hasChildNodes())n.parentNode.insertBefore(n.firstChild, n);
        n = n.parentNode.removeChild(n).nextHighlight;
      }

      var record = n;
      node.nextHighlight = null;  //break chain so loop can exit

      n = record.firstNode;
      while(n instanceof HTMLSpanElement && n.hasAttribute("highlighter")) {
        while(n.hasChildNodes())n.parentNode.insertBefore(n.firstChild, n);
        n = n.parentNode.removeChild(n).nextHighlight;
      }

      record.firstNode = null;
      record.lastNode.nextHighlight = null;
      record.lastNode = null;

      //everything below modifies unsafe objects on the webpage. Do this last.

      var hi = node.ownerDocument.defaultView.highlighterInfo;
      for(var i=hi.highlights.length-1; i>=0; i--) {
        if(hi.highlights[i] == record) {
          hi.highlights.splice(i, 1);
          break;
        }
      }
      hi.dirty = true;
    }
  },

  changeHighlightColor: function(node, fgcolor, bgcolor) {
    var n = node;
    while(n instanceof HTMLSpanElement && n.hasAttribute("highlighter")) {
      n.setAttribute("highlighter", bgcolor);
      n.style.backgroundColor = bgcolor;
      n.style.color = fgcolor;
      n = n.nextHighlight;
    }

    n = n.firstNode;  //firstNode is defined in highlights[i].firstNode
    while(n instanceof HTMLSpanElement && n.hasAttribute("highlighter") && n != node) {
      n.setAttribute("highlighter", bgcolor);
      n.style.backgroundColor = bgcolor;
      n.style.color = fgcolor;
      n = n.nextHighlight;
    }

    node.ownerDocument.defaultView.highlighterInfo.dirty = true;
  },

  gatherText: function(record, limit) {
    var str = "";
    var n = record.firstNode;
    while(n instanceof HTMLSpanElement) {
      var n2 = n.firstChild;
      while(n2 instanceof Text) {
        str += n2.nodeValue;
        if(limit && str.length >= limit)return str;
        n2 = n2.nextSibling;
      }
      n = n.nextHighlight;
    }
    return str;
  },

  copyAllHighlights: function() {
    var doc = document.popupNode.ownerDocument;
    if(!doc.defaultView.highlighterInfo)return;

    var records = doc.defaultView.highlighterInfo.highlights;
    var str = "";
    for(var i=0; i<records.length; i++) {
      str += Highlighter.gatherText(records[i]).replace(/\s+/g, " ") + "\n\n";
    }
    str += new Date().toLocaleString() + "\n" + doc.title + "\n" + doc.location.href;

    Components.classes["@mozilla.org/widget/clipboardhelper;1"]
      .getService(Components.interfaces.nsIClipboardHelper)
      .copyString(str);
  },

///////////////////////////////////////////////////////////////////////////////

  toggleInstantHighlight: function(forceOff) {
    var win = document.commandDispatcher.focusedWindow;
    var sel = win.getSelection();
    if(!forceOff && sel.rangeCount > 0 && !sel.getRangeAt(0).collapsed) {
      var c = Highlighter.lastColors;
      Highlighter.highlightRange(win, c.fgcolor, c.bgcolor);
      return;
    }

    if(forceOff || Highlighter.autoHighlightMode) {
      Highlighter.autoHighlightMode = false;
      document.getElementById("instantHighlighter").setAttribute("checked", "false");
    } else {
      Highlighter.autoHighlightMode = true;
      document.getElementById("instantHighlighter").setAttribute("checked", "true");
    }
  },

  autoHighlight: function(event) {
    if(Highlighter.autoHighlightMode) {
      var node = event.target;

      if(node instanceof HTMLElement) {
        var win = node.ownerDocument.defaultView;
        var c = Highlighter.lastColors;

        Highlighter.highlightRange(win, c.fgcolor, c.bgcolor);
      }
    }
  },

  toolbarHighlightDoc: function(context) {
    var fgcolor = Highlighter.lastColors.fgcolor = context.getAttribute("fgcolor");
    var bgcolor = Highlighter.lastColors.bgcolor = context.getAttribute("bgcolor");

    var win = document.commandDispatcher.focusedWindow;
    var sel = win.getSelection();
    if(sel.rangeCount > 0 && !sel.getRangeAt(0).collapsed) {
      Highlighter.highlightRange(win, fgcolor, bgcolor);
    }
  },

  ensureUnloadWarning: function(win) {
    if(win.document.body.hasAttribute("onbeforeonload"))return;
    win.document.body.setAttribute("onbeforeunload" ,"return 1;");
  },

///////////////////////////////////////////////////////////////////////////////

  ensureHighlighterView: function(win) {
    if(!win.highlighterView) {
      setTimeout(function(w) {
        Highlighter.ensureHighlighterView0(w);
      }, 0, win);
    }
  },

  ensureHighlighterView0: function(win) {
    if(win.highlighterView)return;
    win.highlighterView = {lastPos:0};

    var doc = win.document;
    var sheet = doc.createElement("STYLE");
    sheet.setAttribute("type", "text/css");
    doc.body.appendChild(sheet);
    sheet = sheet.sheet;
    sheet.insertRule(".hl_holder {position:fixed; right:1%; font-size:9pt}", sheet.length);
    sheet.insertRule("#hl_t_holder {top:1%}", sheet.length);
    sheet.insertRule("#hl_b_holder {bottom:1%}", sheet.length);
    sheet.insertRule("#hl_m_holder {display:none}", sheet.length);
    sheet.insertRule(".hl_holder>a {margin:0 1px; padding:0 0.2em; text-decoration:none}", sheet.length);
    sheet.insertRule("@media print {}", sheet.length);
    sheet.cssRules.item(sheet.length-1).insertRule(".hl_holder {display:none}", 0);

    var frag = doc.createDocumentFragment();

    var topHolder = doc.createElement("DIV");
    topHolder.setAttribute("id", "hl_t_holder");
    topHolder.setAttribute("class", "hl_holder");
    frag.appendChild(topHolder);

    var midHolder = doc.createElement("DIV");
    midHolder.setAttribute("id", "hl_m_holder");
    midHolder.setAttribute("class", "hl_holder");
    frag.appendChild(midHolder);

    var btmHolder = doc.createElement("DIV");
    btmHolder.setAttribute("id", "hl_b_holder");
    btmHolder.setAttribute("class", "hl_holder");
    frag.appendChild(btmHolder);

    doc.body.appendChild(frag);

    const _buildItem = function(record) {
      var item = doc.createElement("A");
      item.setAttribute("href", "javascript:scrollTo(pageXOffset,"+record.offsetY+");");
      item.offsetY = record.offsetY;
      item.setAttribute("title", Highlighter.gatherText(record, 80).replace(/\s+/g, " "));
      item.appendChild(doc.createTextNode(decodeURI("%C2%B6")));
      return item;
    };

    const _onScroll = function() {
      var posTop = win.pageYOffset;

      if(!win.highlighterInfo ||
          win.highlighterInfo.highlights.length == 0 ||
          win.highlighterView.lastPos == posTop)
        return;

      win.highlighterView.lastPos = posTop;
      var posBtm = posTop + win.innerHeight;

      if(win.highlighterInfo.dirty) {
        var records = win.highlighterInfo.highlights;
        while(midHolder.hasChildNodes())topHolder.appendChild(midHolder.firstChild);
        while(btmHolder.hasChildNodes())topHolder.appendChild(btmHolder.firstChild);

        for(var i=0; i<records.length; i++) {
          if(!records[i].view)records[i].view = _buildItem(records[i]);
          records[i].view.style.backgroundColor = records[i].firstNode.style.backgroundColor;
          midHolder.appendChild(records[i].view);
        }

        while(topHolder.hasChildNodes())topHolder.removeChild(topHolder.firstChild);
        win.highlighterInfo.dirty = false;
      }

      while(midHolder.lastChild && midHolder.lastChild.offsetY > posBtm)
        btmHolder.insertBefore(midHolder.lastChild, btmHolder.firstChild);
      while(topHolder.lastChild && topHolder.lastChild.offsetY > posBtm)
        btmHolder.insertBefore(topHolder.lastChild, btmHolder.firstChild);
      while(topHolder.lastChild && topHolder.lastChild.offsetY >= posTop)
        midHolder.insertBefore(topHolder.lastChild, midHolder.firstChild);

      while(midHolder.firstChild && midHolder.firstChild.offsetY < posTop)
        topHolder.appendChild(midHolder.firstChild);
      while(btmHolder.firstChild && btmHolder.firstChild.offsetY < posTop)
        topHolder.appendChild(btmHolder.firstChild);
      while(btmHolder.firstChild && btmHolder.firstChild.offsetY <= posBtm)
        midHolder.appendChild(btmHolder.firstChild);
    };

    win.setInterval(_onScroll, 1000);
  }
}

window.addEventListener("load", Highlighter.startup, false);
window.addEventListener("unload", Highlighter.destroy, false);